import logging

from dbtsl.api.shared.query_params import GroupByParam
from mcp.server.fastmcp import FastMCP

from dbt_mcp.config.config import SemanticLayerConfig
from dbt_mcp.prompts.prompts import get_prompt
from dbt_mcp.semantic_layer.client import get_semantic_layer_fetcher
from dbt_mcp.semantic_layer.types import (
    DimensionToolResponse,
    EntityToolResponse,
    MetricToolResponse,
    OrderByParam,
    QueryMetricsSuccess,
)

logger = logging.getLogger(__name__)


def register_sl_tools(dbt_mcp: FastMCP, config: SemanticLayerConfig) -> None:
    semantic_layer_fetcher = get_semantic_layer_fetcher(config)

    @dbt_mcp.tool(description=get_prompt("semantic_layer/list_metrics"))
    def list_metrics() -> list[MetricToolResponse] | str:
        return semantic_layer_fetcher.list_metrics()

    @dbt_mcp.tool(description=get_prompt("semantic_layer/get_dimensions"))
    def get_dimensions(metrics: list[str]) -> list[DimensionToolResponse] | str:
        return semantic_layer_fetcher.get_dimensions(metrics=metrics)

    @dbt_mcp.tool(description=get_prompt("semantic_layer/get_entities"))
    def get_entities(metrics: list[str]) -> list[EntityToolResponse] | str:
        return semantic_layer_fetcher.get_entities(metrics=metrics)

    @dbt_mcp.tool(description=get_prompt("semantic_layer/query_metrics"))
    def query_metrics(
        metrics: list[str],
        group_by: list[GroupByParam] | None = None,
        order_by: list[OrderByParam] | None = None,
        where: str | None = None,
        limit: int | None = None,
    ) -> str:
        result = semantic_layer_fetcher.query_metrics(
            metrics=metrics,
            group_by=group_by,
            order_by=order_by,
            where=where,
            limit=limit,
        )
        if isinstance(result, QueryMetricsSuccess):
            return result.result
        else:
            return result.error



import os
import subprocess
from collections.abc import Iterable

from mcp.server.fastmcp import FastMCP
from pydantic import Field

from dbt_mcp.config.config import DbtCliConfig
from dbt_mcp.prompts.prompts import get_prompt


def register_dbt_cli_tools(dbt_mcp: FastMCP, config: DbtCliConfig) -> None:
    def _run_dbt_command(
        command: list[str],
        selector: str | None = None,
        timeout: int | None = None,
        resource_type: list[str] | None = None,
    ) -> str:
        # Commands that should always be quiet to reduce output verbosity
        verbose_commands = ["build", "compile", "docs", "parse", "run", "test", "list"]

        if selector:
            selector_params = str(selector).split(" ")
            command = command + ["--select"] + selector_params

        if isinstance(resource_type, Iterable):
            command = command + ["--resource-type"] + resource_type

        full_command = command.copy()
        # Add --quiet flag to specific commands to reduce context window usage
        if len(full_command) > 0 and full_command[0] in verbose_commands:
            main_command = full_command[0]
            command_args = full_command[1:] if len(full_command) > 1 else []
            full_command = [main_command, "--quiet", *command_args]

        # We change the path only if this is an absolute path, otherwise we can have
        # problems with relative paths applied multiple times as DBT_PROJECT_DIR
        # is applied to dbt Core and Fusion as well (but not the dbt Cloud CLI)
        cwd_path = config.project_dir if os.path.isabs(config.project_dir) else None

        process = subprocess.Popen(
            args=[config.dbt_path, *full_command],
            cwd=cwd_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )
        output, _ = process.communicate(timeout=timeout)
        return output or "OK"

    @dbt_mcp.tool(description=get_prompt("dbt_cli/build"))
    def build(
        selector: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/selectors")
        ),
    ) -> str:
        return _run_dbt_command(["build"], selector)

    @dbt_mcp.tool(description=get_prompt("dbt_cli/compile"))
    def compile() -> str:
        return _run_dbt_command(["compile"])

    @dbt_mcp.tool(description=get_prompt("dbt_cli/docs"))
    def docs() -> str:
        return _run_dbt_command(["docs", "generate"])

    @dbt_mcp.tool(name="list", description=get_prompt("dbt_cli/list"))
    def ls(
        selector: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/selectors")
        ),
        resource_type: list[str] | None = Field(
            default=None,
            description=get_prompt("dbt_cli/args/resource_type"),
        ),
    ) -> str:
        try:
            return _run_dbt_command(
                ["list"], selector, timeout=10, resource_type=resource_type
            )
        except subprocess.TimeoutExpired:
            return (
                "Timeout: dbt list command took too long to complete. "
                + "Try using a more specific selector to narrow down the list of models."
            )

    @dbt_mcp.tool(description=get_prompt("dbt_cli/parse"))
    def parse() -> str:
        return _run_dbt_command(["parse"])

    @dbt_mcp.tool(description=get_prompt("dbt_cli/run"))
    def run(
        selector: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/selectors")
        ),
    ) -> str:
        return _run_dbt_command(["run"], selector)

    @dbt_mcp.tool(description=get_prompt("dbt_cli/test"))
    def test(
        selector: str | None = Field(
            default=None, description=get_prompt("dbt_cli/args/selectors")
        ),
    ) -> str:
        return _run_dbt_command(["test"], selector)

    @dbt_mcp.tool(description=get_prompt("dbt_cli/show"))
    def show(
        sql_query: str = Field(description=get_prompt("dbt_cli/args/sql_query")),
        limit: int | None = Field(
            default=None, description=get_prompt("dbt_cli/args/limit")
        ),
    ) -> str:
        args = ["show", "--inline", sql_query, "--favor-state"]
        # This is quite crude, but it should be okay for now
        # until we have a dbt Fusion integration.
        cli_limit = None
        if "limit" in sql_query.lower():
            # When --limit=-1, dbt won't apply a separate limit.
            cli_limit = -1
        elif limit:
            # This can be problematic if the LLM provides
            # a SQL limit and a `limit` argument. However, preferencing the limit
            # in the SQL query leads to a better experience when the LLM
            # makes that mistake.
            cli_limit = limit
        if cli_limit is not None:
            args.extend(["--limit", str(cli_limit)])
        args.extend(["--output", "json"])
        return _run_dbt_command(args)



import logging

from mcp.server.fastmcp import FastMCP

from dbt_mcp.config.config import DiscoveryConfig
from dbt_mcp.discovery.client import MetadataAPIClient, ModelsFetcher
from dbt_mcp.prompts.prompts import get_prompt

logger = logging.getLogger(__name__)


def register_discovery_tools(dbt_mcp: FastMCP, config: DiscoveryConfig) -> None:
    api_client = MetadataAPIClient(
        url=config.url,
        headers=config.headers,
    )
    models_fetcher = ModelsFetcher(
        api_client=api_client, environment_id=config.environment_id
    )

    @dbt_mcp.tool(description=get_prompt("discovery/get_mart_models"))
    def get_mart_models() -> list[dict] | str:
        mart_models = models_fetcher.fetch_models(
            model_filter={"modelingLayer": "marts"}
        )
        return [m for m in mart_models if m["name"] != "metricflow_time_spine"]

    @dbt_mcp.tool(description=get_prompt("discovery/get_all_models"))
    def get_all_models() -> list[dict] | str:
        return models_fetcher.fetch_models()

    @dbt_mcp.tool(description=get_prompt("discovery/get_model_details"))
    def get_model_details(model_name: str, unique_id: str | None = None) -> dict | str:
        return models_fetcher.fetch_model_details(model_name, unique_id)

    @dbt_mcp.tool(description=get_prompt("discovery/get_model_parents"))
    def get_model_parents(
        model_name: str, unique_id: str | None = None
    ) -> list[dict] | str:
        return models_fetcher.fetch_model_parents(model_name, unique_id)

    @dbt_mcp.tool(description=get_prompt("discovery/get_model_children"))
    def get_model_children(
        model_name: str, unique_id: str | None = None
    ) -> list[dict] | str:
        return models_fetcher.fetch_model_children(model_name, unique_id)



import os
from dataclasses import dataclass
from pathlib import Path

import yaml
from dotenv import load_dotenv


@dataclass
class TrackingConfig:
    host: str | None
    multicell_account_prefix: str | None
    prod_environment_id: int | None
    dev_environment_id: int | None
    dbt_cloud_user_id: int | None
    local_user_id: str | None


@dataclass
class SemanticLayerConfig:
    multicell_account_prefix: str | None
    host: str
    prod_environment_id: int
    service_token: str


@dataclass
class DiscoveryConfig:
    url: str
    headers: dict[str, str]
    environment_id: int


@dataclass
class DbtCliConfig:
    project_dir: str
    dbt_path: str


@dataclass
class RemoteConfig:
    multicell_account_prefix: str | None
    host: str
    user_id: int
    dev_environment_id: int
    prod_environment_id: int
    token: str


@dataclass
class Config:
    tracking_config: TrackingConfig
    remote_config: RemoteConfig | None
    dbt_cli_config: DbtCliConfig | None
    discovery_config: DiscoveryConfig | None
    semantic_layer_config: SemanticLayerConfig | None


def load_config() -> Config:
    load_dotenv()

    host = os.environ.get("DBT_HOST")
    cursor_host = os.environ.get("DBT_MCP_HOST")
    prod_environment_id = os.environ.get("DBT_PROD_ENV_ID")
    legacy_prod_environment_id = os.environ.get("DBT_ENV_ID")
    dev_environment_id = os.environ.get("DBT_DEV_ENV_ID")
    user_id = os.environ.get("DBT_USER_ID")
    token = os.environ.get("DBT_TOKEN")
    project_dir = os.environ.get("DBT_PROJECT_DIR")
    dbt_path = os.environ.get("DBT_PATH", "dbt")
    disable_dbt_cli = os.environ.get("DISABLE_DBT_CLI", "false") == "true"
    disable_semantic_layer = os.environ.get("DISABLE_SEMANTIC_LAYER", "false") == "true"
    disable_discovery = os.environ.get("DISABLE_DISCOVERY", "false") == "true"
    disable_remote = os.environ.get("DISABLE_REMOTE", "true") == "true"
    multicell_account_prefix = os.environ.get("MULTICELL_ACCOUNT_PREFIX", None)

    # set default warn error options if not provided
    if os.environ.get("DBT_WARN_ERROR_OPTIONS") is None:
        warn_error_options = '{"error": ["NoNodesForSelectionCriteria"]}'
        os.environ["DBT_WARN_ERROR_OPTIONS"] = warn_error_options

    # Devon: I messed up and uploaded the wrong
    # env var here https://docs.cursor.com/tools.
    # So, we are supporting this alias now.
    actual_host = host or cursor_host

    errors = []
    if not disable_semantic_layer or not disable_discovery or not disable_remote:
        if not actual_host:
            errors.append(
                "DBT_HOST environment variable is required when semantic layer, discovery, or remote tools are enabled."
            )
        if not prod_environment_id and not legacy_prod_environment_id:
            errors.append(
                "DBT_PROD_ENV_ID environment variable is required when semantic layer, discovery, or remote tools are enabled."
            )
        if not token:
            errors.append(
                "DBT_TOKEN environment variable is required when semantic layer, discovery, or remote tools are enabled."
            )
        if actual_host and (
            actual_host.startswith("metadata")
            or actual_host.startswith("semantic-layer")
        ):
            errors.append(
                "DBT_HOST must not start with 'metadata' or 'semantic-layer'."
            )
    if not disable_remote:
        if not dev_environment_id:
            errors.append(
                "DBT_DEV_ENV_ID environment variable is required when remote tools are enabled."
            )
        if not user_id:
            errors.append(
                "DBT_USER_ID environment variable is required when remote tools are enabled."
            )
    if not disable_dbt_cli:
        if not project_dir:
            errors.append(
                "DBT_PROJECT_DIR environment variable is required when dbt CLI tools are enabled."
            )
        if not dbt_path:
            errors.append(
                "DBT_PATH environment variable is required when dbt CLI tools are enabled."
            )

    if errors:
        raise ValueError("Errors found in configuration:\n\n" + "\n".join(errors))

    # Allowing for configuration with legacy DBT_ENV_ID environment variable
    actual_prod_environment_id = (
        int(prod_environment_id)
        if prod_environment_id
        else int(legacy_prod_environment_id)
        if legacy_prod_environment_id
        else None
    )

    remote_config = None
    if (
        not disable_remote
        and user_id
        and token
        and dev_environment_id
        and actual_prod_environment_id
        and actual_host
    ):
        remote_config = RemoteConfig(
            multicell_account_prefix=multicell_account_prefix,
            user_id=int(user_id),
            token=token,
            dev_environment_id=int(dev_environment_id),
            prod_environment_id=actual_prod_environment_id,
            host=actual_host,
        )

    dbt_cli_config = None
    if not disable_dbt_cli and project_dir and dbt_path:
        dbt_cli_config = DbtCliConfig(
            project_dir=project_dir,
            dbt_path=dbt_path,
        )

    discovery_config = None
    if not disable_discovery and actual_host and actual_prod_environment_id and token:
        if multicell_account_prefix:
            url = f"https://{multicell_account_prefix}.metadata.{actual_host}/graphql"
        else:
            url = f"https://metadata.{actual_host}/graphql"
        discovery_config = DiscoveryConfig(
            url=url,
            headers={
                "Authorization": f"Bearer {token}",
                "Content-Type": "application/json",
            },
            environment_id=actual_prod_environment_id,
        )

    semantic_layer_config = None
    if (
        not disable_semantic_layer
        and actual_host
        and actual_prod_environment_id
        and token
    ):
        semantic_layer_config = SemanticLayerConfig(
            multicell_account_prefix=multicell_account_prefix,
            host=actual_host,
            prod_environment_id=actual_prod_environment_id,
            service_token=token,
        )

    local_user_id = None
    try:
        home = os.environ.get("HOME")
        user_path = Path(f"{home}/.dbt/.user.yml")
        if home and user_path.exists():
            with open(user_path) as f:
                local_user_id = yaml.safe_load(f).get("id")
    except Exception:
        pass

    return Config(
        tracking_config=TrackingConfig(
            host=actual_host,
            multicell_account_prefix=multicell_account_prefix,
            prod_environment_id=actual_prod_environment_id,
            dev_environment_id=int(dev_environment_id) if dev_environment_id else None,
            dbt_cloud_user_id=int(user_id) if user_id else None,
            local_user_id=local_user_id,
        ),
        remote_config=remote_config,
        dbt_cli_config=dbt_cli_config,
        discovery_config=discovery_config,
        semantic_layer_config=semantic_layer_config,
    )
